Module PSlib
!===============================================================================
!     PSlib module contains subroutines useful for pseudo-spectral methods
!-------------------------------------------------------------------------------
implicit none

private
public :: ffork, wavnum, FWAVE, difx, difx2d, difz2d

Contains
!===============================================================================
subroutine ffork(lx, cx, signi) 
!-------------------------------------------------------------------------------
      implicit none
      integer , intent(in) :: lx 
      real , intent(in) :: signi 
      complex , intent(inout) :: cx(*) 
!-----------------------------------------------
      integer :: j, i, m, l, istep 
      real :: pi, sc 
      complex :: carg, cw, ctemp 
!-----------------------------------------------
      pi = 3.14159265 
      j  = 1 
      sc = 1./float ( lx )  
      sc = sqrt ( sc )  
      do i = 1, lx 
         if (i <= j) then 
            ctemp = cx(j)*sc 
            cx(j) = cx(i)*sc 
            cx(i) = ctemp 
         endif 
         m = lx/2 
         if (j <= m) go to 100 
         j = j - m 
         m = m/2 
         do while ( m >= 1 ) 
            if (j <= m) go to 100 
            j = j - m 
            m = m/2 
         end do 
  100    continue 
         j = j + m 
      end do 
      l = 1 
  110 continue 
      istep = 2*l 
      do m = 1, l 
         carg = cmplx(0.,1.) * ( pi*signi*float(m - 1) ) /float (l)  
         cw   = cexp ( carg )  
         do i = m, lx, istep 
            ctemp   = cw*cx(i+l) 
            cx(i+l) = cx(i) - ctemp 
            cx(i)   = cx(i) + ctemp 
         end do 
      end do 
      l = istep 
      if (l < lx) go to 110 
      return  
end subroutine ffork 

!===============================================================================
subroutine wavnum(ck,n,d,iord)
!-------------------------------------------------------------------------------
implicit none

integer :: i, iord, n
real :: d, pi
complex, dimension (:) :: ck

pi = 4.*atan(1.)

do i=1, n/2+1
   ck(i) = (0.,1.)*2.*pi/(n*d)*(i-1)
end do

do i=n/2+2, n
   ck(i) = -(0.,1.)*2.*pi/(n*d)*(n-(i-1))
end do

if (iord > 1) then
   do i=1, n
      ck(i) = ck(i)**iord
   end do
end if

end subroutine wavnum

!===============================================================================
real  FUNCTION FWAVE(TIME,EFHZ)
!-------------------------------------------------------------------------------
! --- Ricker wavelet ---
!
  implicit none
  SAVE

  integer, save :: ienter=0
  real :: time, efhz
  real :: pi, pi2, agauss, tcut, s

  IF(IENTER.EQ.0) THEN
     IENTER=1
     PI=3.14159265
     PI2=2.*PI
     AGAUSS=0.5*EFHZ
     TCUT=1.5/AGAUSS
  END IF
  s=(time-tcut)*agauss
  fwave=0.
  if(abs(s).lt.4.) fwave=exp(-2.*s*s)*cos(pi2*s)
  return
end FUNCTION FWAVE

!===============================================================================
subroutine difx(vin, vout, ckx, nx)
!-------------------------------------------------------------------------------
implicit none
save icall, cx
integer :: i, icall=0, nx
real, dimension (:) :: vin, vout
complex, dimension (:) :: ckx

complex, allocatable, dimension (:) :: cx

if (icall == 0) then
  icall = 1
!  print*,icall
  allocate(cx(nx))
endif

do i=1, nx
!   print*,vin(i)
   cx(i) = vin(i)
end do

call ffork(nx,cx,-1.)

do i=1, nx
   cx(i) = cx(i)*ckx(i)
end do

call ffork(nx,cx,1.)

do i=1, nx
   vout(i) = real(cx(i))
end do

end subroutine difx

!===============================================================================
subroutine difx2d(vin, vout, ckx, nx, nz)
!-------------------------------------------------------------------------------
implicit none
save icall, cx
integer :: i, k, icall=0, nx, nz
real, dimension (:,:) :: vin, vout
complex, dimension (:) :: ckx

complex, allocatable, dimension (:) :: cx

if (icall == 0) then
  icall = 1
  print*,icall
  allocate(cx(nx))
endif


do k=1, nz
   do i=1, nx
      cx(i) = vin(i,k)
   end do

   call ffork(nx,cx,-1.)

   do i=1, nx
      cx(i) = cx(i)*ckx(i)
   end do

   call ffork(nx,cx,1.)

   do i=1, nx
      vout(i,k) = real(cx(i))
   end do
end do

end subroutine difx2d

!===============================================================================
subroutine difz2d(vin, vout, ckz, nx, nz)
!-------------------------------------------------------------------------------
implicit none
save icall, cx
integer :: i, k, icall=0, nx, nz
real, dimension (:,:) :: vin, vout
complex, dimension (:) :: ckz

complex, allocatable, dimension (:) :: cx

if (icall == 0) then
  icall = 1
  allocate(cx(nz))
  print*,icall
endif


do i=1, nx
   do k=1, nz
      cx(k) = vin(i,k)
   end do

   call ffork(nz,cx,-1.)

   do k=1, nz
      cx(k) = cx(k)*ckz(k)
   end do

   call ffork(nz,cx,1.)

   do k=1, nz
      vout(i,k) = real(cx(k))
   end do
end do

end subroutine difz2d


End Module PSlib    
