program ac2d
!
! acoustic 2-D Fourier modelling program
!
USE PSlib 
USE SVlib

! declaration of variables
implicit none

integer, parameter :: SU=1, SV=2  !(Seismic Unix / SismoVi)
integer :: i, isx, isz, k, l, nt, nx, nz, export_opt
real :: amx, arg, c, cmax, dx, dz, efhz, pi, dt, dt2, time, tmax, tmp
real, allocatable, dimension (:,:) :: p, pp, c2, a1, a2
complex, allocatable, dimension (:) :: ckx, ckz

! initialization of variables

nx = 128
nz = 128
pi = 4.*atan(1.)
dx = 10.
dz = 10.
efhz = 100.
c = 2000.
isx = 32
isz = 56
tmax = 0.25
!-------------------------------------------------------
 export_opt = SV    ! SU for Seismic Unix / SV if SismoVi is used  !!!!!!!!!!!!
!-------------------------------------------------------

allocate(p(nx,nz), pp(nx,nz), c2(nx,nz), a1(nx,nz), a2(nx,nz), &
         ckx(nx), ckz(nz))

call wavnum(ckx,nx,dx,2)
call wavnum(ckz,nz,dz,2)

do k=1, nz
   do i=1, nx
      p(i,k) = 0.
      pp(i,k) = 0.
      c2(i,k) = c*c
   end do
end do

! put second material
!c = 4000.
!do k=65, nz
!   do i=1, nx
!      c2(i,k) = c*c
!   end do
!end do

cmax=0.
do k=1, nz
   do i=1,nx
      cmax = amax1(cmax,sqrt(c2(i,k)))
   end do
end do

! calculate dt from stability criterion, multiply by 0.2 for less dispersion
dt = sqrt(2.)*dx/(pi*cmax) * 0.2
dt2 = dt * dt
nt = tmax/dt

! loop over time steps

do l=1, nt
   time = l * dt

   call difx2d(p,a1,ckx,nx,nz)

   call difz2d(p,a2,ckz,nx,nz)

   do k=1, nz
      do i=1,nx
         a1(i,k) = (a1(i,k) + a2(i,k)) * c2(i,k)
      end do
   end do

! add source term
   a1(isx,isz) = a1(isx,isz) + FWAVE(TIME,EFHZ)

! do time integration
   do k=1, nz
      do i=1,nx
         pp(i,k) = -pp(i,k) + 2.*p(i,k) + dt2 * a1(i,k)
      end do
   end do

! rearrange arrays p^(n+1), p^(n), p^(n-1)
   do k=1, nz
      do i=1,nx
         tmp = pp(i,k)
         pp(i,k) = p(i,k)
         p(i,k) = tmp
      end do
   end do
   
! calculate maximum
   amx = 0.
   do k=1, nz
      do i=1, nx
         amx = amax1(amx,abs(p(i,k)))
      end do
   end do
   print*,l,amx

! save 10 snapshots, every (nt/10) time steps
   if (mod(l,nt/10) == 0) then
      select case (export_opt)
        case (SU)        ! export data for SU visualization
            do i=1, nx
               write(10) p(i,:)
            end do
        case (SV)        ! export data for SismoVi visualization
            call export_snap(p)
        case default
            STOP "Export error"
      end select
   end if

end do ! end of loop over time steps

stop "<> ac2d <>"
Contains
!------------------------------------------------------- 
 Subroutine export_snap(pdata)
   implicit none
   real :: pdata(:,:)
   integer, save :: snap=0
   character(len=3) :: sn
   character(len=6) :: tn

   snap = snap + 1
   write(sn,"(I3.3)") snap
   write(tn,"(f6.3)") time
   Call SismoViExport(BEGIN, fname='ac2d.snap'//sn, title='Acoustic 2D -'//tn//' s',  &
                       xlabel='Space',  ylabel='Space',                        &
                       traces= SIZE(pdata, DIM=2), samples=SIZE(pdata, DIM=1), &
                       tracestep=dz,  timestep=dx )
   Call SismoViExport( array=pdata ) 
   Call SismoViExport(FINIS)
      
 End subroutine export_snap

end program ac2d
